/*----------------------------------------------*/
/*						*/
/*	Lexical Analyser			*/
/*		for reading CBR case records	*/
/*						*/
/*						*/
/* Modified for version 2			*/
/* Modified for version 3			*/
/* Modified for version 9			*/
/* Modified for version 10			*/
/* Modified for version 11			*/
/* Modified for version 12			*/
/* Modified for version 13			*/
/* Modified for version 14			*/
/* Modified for version 15			*/
/* Modified for version 16			*/
/* Modified for version 17			*/
/* Modified for version 18			*/
/* Modified for version 23			*/
/* Modified for version 24			*/
/* Modified for version 28			*/
/* Modified for version 29			*/
/* Modified for version 31			*/
/* Modified for version 32			*/
/* Modified for version 33			*/
/* Modified for version 35			*/
/* Modified for version 36			*/
/* Modified for version 38			*/
/* Modified for version 39			*/
/* Modified for version 43			*/
/* Modified for version 45			*/
/* Modified for version 46			*/
/* Modified for version 49			*/
/*----------------------------------------------*/


void skipwhitespace()
{

 /* skip any white space */
 
 while (ch == ' ' || ch == '\t' || ch == '\n')
   {
    if (ch == '\n') lines++; /* count the number of newlines */
    ch = getc(fp);
   }

 if (ch == '~')
    skipcomments();
}

void skipcomments()
{

 /* skip past comments */
 
 if (ch == '~')   
   { 
    while (ch != EOF && ch != '\n') 
      { 
       ch = getc(fp);
      }  
   }

 if (ch == ' ' || ch == '\t' || ch == '\n')
    skipwhitespace();
}


int yylex()

{

int n; /* loop count */
float value, frac, dec; /* values for calculating real value */


 /* if this is the first time that the lexical analyser has been
    called then read in the first character */

 if (START == TRUE)
   {
    ch = getc(fp);
    START = FALSE;
   }


 /* skip any white space */

 skipwhitespace();

 /* if ch = EOF then we have reached the end of file */

 if ( ch == EOF )
    return 0;

 if (ch == '\'') /* lexeme is a string */
   {
    ch = getc(fp); /* read in first character of string */

    yylval.strval = createstringnode(); /*create node to contain string */
    n = 0;
    while (n < STRBUFSIZE-1 && ch != '\'')
      {
       if (ch == '\n')
         {
          yyerror("newline character in string");
         }
       *(yylval.strval+n) = ch; /* copy character across to buffer */
       ch = getc(fp); /* read next character */
       n++; /* increment count */
      }

    if (n == STRBUFSIZE-1 && ch != '\'')
      {
       fprintf(stderr,"Error - STRBUFSIZE not big enough\n");
       exit(1);
      }

    *(yylval.strval+n) = NULL; /* terminate string */
    ch = getc(fp); /* read first character after the second ' */
    return STRING;
   }

 else if (ch >= '0' && ch <= '9') /* lexeme is a real */
   {
    value = frac = 0.0; /* set value to 0 */

    while (ch >= '0' && ch <= '9') /* read in whole part */
      {
       value = (10.0 * value) + (float) (ch - '0');
       ch = getc(fp);
      }

    if (ch == '.') /* number has fractional part */
      {
       ch = getc(fp);
       dec = 0.1;
       while (ch >= '0' && ch <= '9') /* read in fractional part */
         {
          frac += dec * (float) (ch - '0');
          dec /= 10.0;
          ch = getc(fp);
         }
      }

    yylval.realval = (value + frac); /* copy real number value to yylval */
    return REALNUM;
   }

 else if (ch == ':')
   {
    ch = getc(fp);
    return ':';
   }

 else if (ch == '[')
   {
    ch = getc(fp);
    return '[';
   }

 else if (ch == ']')
   {
    ch = getc(fp);
    return ']';
   }

 else if (ch == ';')
   {
    ch = getc(fp);
    return ';';
   }

 else if (ch == '=')
   {
    ch = getc(fp);
    yylval.comptype = '=';
    return '=';
   }

 else if (ch == '<')
   {
    ch = getc(fp);
    if (ch == '=')
      {
       ch = getc(fp);
       yylval.comptype = LESSOREQUAL;
       return LESSOREQUAL;
      }
    else
      {
       yylval.comptype = '<';
       return '<';
      }
   }

 else if (ch == '>')
   {
    ch = getc(fp);
    if (ch == '=')
      {
       ch = getc(fp);
       yylval.comptype = MOREOREQUAL;
       return MOREOREQUAL;
      }
    else
      {
       yylval.comptype = '>';
       return '>';
      }
   }

 else if (ch == '!')
   {
    ch = getc(fp);
    if (ch == '=')
      {
       ch = getc(fp);
       yylval.comptype = NOTEQUAL;
       return NOTEQUAL;
      }
   }

 else if (ch == '(')
   {
    ch = getc(fp);
    return '(';
   }

 else if (ch == ')')
   {
    ch = getc(fp);
    return ')';
   }

 else if (ch == '+')
   {
    ch = getc(fp);
    return '+';
   }

else if (ch == '-')
   {
    ch = getc(fp);
    return '-';
   }

else if (ch == '*')
   {
    ch = getc(fp);
    return '*';
   }

else if (ch == '/')
   {
    ch = getc(fp);
    return '/';
   }

else if (ch == '^')
   {
    ch = getc(fp);
    return '^';
   }
 else if (ch == ',')
   {
    ch = getc(fp);
    return ',';
   }

 else if ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z')) /* lexeme is identifier or keyword*/
   {
    yylval.strval = createstringnode(); /* create node to contain lexeme */
    n = 0;
    while ((ch >= 'a' && ch <= 'z') ||
           (ch >= 'A' && ch <= 'Z') ||
           (ch >= '0' && ch <= '9') ||
           (ch == '_'))
      {
       if (n == STRBUFSIZE-1)
         {
          yyerror("lexeme is too long");
         }
       *(yylval.strval+n) = ch;
       ch = getc(fp);
       n++;
      }
    *(yylval.strval+n) = NULL; /* terminate identifier/keyword lexeme */

/* return the appropriate keyword symbol if it matches a keyword */

    if (strcmp(yylval.strval, "abstraction") == 0)
      {
       free(yylval.strval);
       return ABSTRACTION;
      }
    else if (strcmp(yylval.strval, "and") == 0)
      {
       free(yylval.strval);
       return AND;
      }
    else if (strcmp(yylval.strval, "case") == 0)
      {
       free(yylval.strval); /* we don't need the buffer any more */
       return CASE;
      }
    else if (strcmp(yylval.strval, "change") == 0)
      {  
       free(yylval.strval);
       return CHANGE;
      }
    else if (strcmp(yylval.strval, "definition") == 0)
      {
       free(yylval.strval);
       return DEFINITION;
      }
    else if (strcmp(yylval.strval, "end") == 0)
      {
       free(yylval.strval);
       return END;
      }
    else if (strcmp(yylval.strval, "evaluate") == 0)
      {
       free(yylval.strval);
       return EVALUATE;
      }
    else if (strcmp(yylval.strval, "extract") == 0)
      {
       free(yylval.strval);
       return EXTRACT;
      }
    else if (strcmp(yylval.strval, "field") == 0)
      {
       free(yylval.strval);
       return FIELD;
      }
    else if (strcmp(yylval.strval, "from") == 0)
      {
       free(yylval.strval);
       return FROM;
      }
    else if (strcmp(yylval.strval, "hd") == 0)
      {
       free(yylval.strval);
       return HD;
      }
    else if (strcmp(yylval.strval, "index") == 0)
      {
       free(yylval.strval);
       return INDEX;
      }
    else if (strcmp(yylval.strval, "instance") == 0)
      {
       free(yylval.strval);
       return INSTANCE;
      }
    else if (strcmp(yylval.strval, "introduction") == 0)
      {
       free(yylval.strval);
       return INTRODUCTION;
      }
    else if (strcmp(yylval.strval, "is") == 0)
      {
       free(yylval.strval);
       return IS;
      }
    else if (strcmp(yylval.strval, "list") == 0)
      {
       free(yylval.strval);
       return LIST;
      }
    else if (strcmp(yylval.strval, "local") == 0)
      {
       free(yylval.strval);
       return LOCAL;
      }
    else if (strcmp(yylval.strval, "modification") == 0)
      {  
       free(yylval.strval);
       return MODIFICATION;
      }
    else if (strcmp(yylval.strval, "not") == 0)
      {
       free(yylval.strval);
       return NOT;
      }
    else if (strcmp(yylval.strval, "number") == 0)
      {
       free(yylval.strval);
       return NUMBER;
      }
    else if (strcmp(yylval.strval, "of") == 0)
      {
       free(yylval.strval);
       return OF;
      }
    else if (strcmp(yylval.strval, "on") == 0)
      {
       free(yylval.strval);
       return ON;
      }
    else if (strcmp(yylval.strval, "or") == 0)
      {
       free(yylval.strval);
       return OR;
      }
    else if (strcmp(yylval.strval, "pr") == 0)
       {
        free(yylval.strval);
        return PR;
       }
    else if (strcmp(yylval.strval, "preprocess") == 0)
       {
        free(yylval.strval);
        return PREPROCESS;
       }
    else if (strcmp(yylval.strval, "prompt") == 0)
       {
        free(yylval.strval);
        return PROMPT;
       }
    else if (strcmp(yylval.strval, "range") == 0)
      {  
       free(yylval.strval);
       return RANGE;
      }
    else if (strcmp(yylval.strval, "repair") == 0)
      {  
       free(yylval.strval);
       return REPAIR;
      }
    else if (strcmp(yylval.strval, "reselect") == 0)
      {
       free(yylval.strval);
       return RESELECT;
      }
    else if (strcmp(yylval.strval, "rule") == 0)
      {  
       free(yylval.strval);
       return RULE;
      }
    else if (strcmp(yylval.strval, "similar") == 0)
      {  
       free(yylval.strval);
       return SIMILAR;
      }
    else if (strcmp(yylval.strval, "solution") == 0)
     {
      free(yylval.strval);
      return SOLUTION;
     }
    else if (strcmp(yylval.strval, "string") == 0)
      {
       free(yylval.strval);
       return KSTRING;
      }
    else if (strcmp(yylval.strval, "then") == 0)
      {
       free(yylval.strval);
       return THEN;
      }
    else if (strcmp(yylval.strval, "tl") == 0)
      {
       free(yylval.strval);
       return TL;
      }
    else if (strcmp(yylval.strval, "to") == 0)
      {  
       free(yylval.strval);
       return TO;
      }
    else if (strcmp(yylval.strval, "type") == 0)
      {
       free(yylval.strval);
       return TYPE;
      }
    else if (strcmp(yylval.strval, "undefined") == 0)
      {  
       free(yylval.strval);
       return UNDEFINED;
      }
    else if (strcmp(yylval.strval, "when") == 0)
      {
       free(yylval.strval);
       return WHEN;
      }
    else if (strcmp(yylval.strval, "weight") == 0)
      {
       free(yylval.strval);
       return WEIGHT;
      }
    else
       return IDENTIFIER; /* otherwise it must be an identifier */
   }
 yyerror("unidentified lexeme");
}

